"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = void 0;

var _ethereumjsUtil = require("ethereumjs-util");

var _merkleTree = _interopRequireDefault(require("./merkleTree"));

var _block = _interopRequireDefault(require("./block"));

var _util = _interopRequireDefault(require("./util"));

var _constants = require("./constants");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _slicedToArray(arr, i) { return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest(); }

function _nonIterableRest() { throw new TypeError("Invalid attempt to destructure non-iterable instance"); }

function _iterableToArrayLimit(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"] != null) _i["return"](); } finally { if (_d) throw _e; } } return _arr; }

function _arrayWithHoles(arr) { if (Array.isArray(arr)) return arr; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

var Period =
/*#__PURE__*/
function () {
  /**
   * 
   * Create a Period object linked to the previous Period via `prevHash` and
   * containing given list of `blocks`.
   * 
   * Optional parameters can be given via `opts` argument. Format as follows:
   * 
   * type PeriodOptions = {
   *   validatorData?: {
   *     slotId: number;
   *     ownerAddr: string | Buffer | number;
   *     casBitmap?: string | Buffer | number; 
   *   };
   *   excludePrevHashFromProof?: Boolean;
   * };
   * 
   * @param {string} prevHash - previous period root hash
   * @param {Array<Block>} blocks - array of blocks to include in the period
   * @param {PeriodOptions} opts - options defined as above. Default {}
   */
  function Period(prevHash, blocks) {
    var _this = this;

    var opts = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    _classCallCheck(this, Period);

    this.prevHash = prevHash;
    this.blockList = [];
    this.blockHashList = [];
    this.usePrev = !opts.excludePrevHashFromProof;

    if (blocks) {
      blocks.forEach(function (block) {
        return _this.addBlock(block);
      });
    }

    var _ref = opts.validatorData || {},
        slotId = _ref.slotId,
        ownerAddr = _ref.ownerAddr,
        casBitmap = _ref.casBitmap;

    if ((slotId || slotId === 0) && ownerAddr) {
      this.setValidatorData(slotId, ownerAddr, casBitmap);
    }
  }

  _createClass(Period, [{
    key: "addBlock",
    value: function addBlock(block) {
      if (this.blockHashList.indexOf(block.hash()) > -1) {
        throw Error('block already contained');
      }

      this.blockList.push(block);
      this.blockHashList.push(block.hash()); // invalidate existing tree
      // todo: optimize not to recalculate the whole tree on every tx added

      this.merkleTree = null;
      return this;
    }
  }, {
    key: "getMerkleTree",
    value: function getMerkleTree() {
      if (this.blockList.length === 0) {
        throw Error('no blocks contained');
      }

      if (!this.merkleTree) {
        var hashBufs = this.blockList.map(function (block) {
          if (block.txList.length === 0) {
            return Buffer.alloc(32, 0);
          }

          return block.getMerkleTree().getRoot();
        });
        this.merkleTree = new _merkleTree["default"](hashBufs);
      }

      return this.merkleTree;
    }
  }, {
    key: "merkleRoot",
    value: function merkleRoot() {
      return this.getMerkleTree().getHexRoot();
    }
  }, {
    key: "setValidatorData",
    value: function setValidatorData(slotId, ownerAddr, casBitmap) {
      this.slotId = slotId;
      this.ownerAddr = (0, _ethereumjsUtil.toBuffer)(ownerAddr);

      if (casBitmap) {
        this.casBitmap = (0, _ethereumjsUtil.toBuffer)(casBitmap);
      }
    } // 

    /**
     * helpful: https://docs.google.com/drawings/d/13oFjua-v_E_yaFYUbralluI-EysgtTaZb_sSgGbxG_A
     * 
     *                          period root
     *                        /             \
     *          consensus root               CAS root  
     *         /             \             /          \
     *   blocks root      meta root     CAS bitmap    validator root
     *                   /        \                         /      \
     *          fees hash    prevPeriodHash         * slotId        0x0
     *                                              * ownerAddr
     */

  }, {
    key: "periodData",
    value: function periodData() {
      if (typeof this.slotId === 'undefined' || !this.ownerAddr) {
        throw Error('period is missing validator data to create period root.');
      } // make validator Root


      var rootBuf = Buffer.alloc(64, 0);
      rootBuf.writeUInt8(this.slotId, 11);
      this.ownerAddr.copy(rootBuf, 12);
      var validatorRoot = (0, _ethereumjsUtil.keccak256)(rootBuf); // make casRoot

      rootBuf = Buffer.alloc(64, 0);

      if (this.casBitmap) {
        this.casBitmap.copy(rootBuf, 0);
      }

      validatorRoot.copy(rootBuf, 32);
      var casRoot = (0, _ethereumjsUtil.keccak256)(rootBuf); // make metaRoot

      var metaRoot = Buffer.alloc(32, 0);

      if (this.usePrev) {
        rootBuf = Buffer.alloc(64, 0);
        (0, _ethereumjsUtil.toBuffer)(this.prevHash).copy(rootBuf, 32);
        metaRoot = (0, _ethereumjsUtil.keccak256)(rootBuf);
      } // make consensusRoot


      var blocksRoot = this.getMerkleTree().getRoot();
      rootBuf = Buffer.alloc(64, 0);
      blocksRoot.copy(rootBuf);
      metaRoot.copy(rootBuf, 32);
      var consensusRoot = (0, _ethereumjsUtil.keccak256)(rootBuf); // make period root

      rootBuf = Buffer.alloc(64, 0);
      consensusRoot.copy(rootBuf);
      casRoot.copy(rootBuf, 32);
      var periodRoot = (0, _ethereumjsUtil.bufferToHex)((0, _ethereumjsUtil.keccak256)(rootBuf));
      return {
        casRoot: casRoot,
        periodRoot: periodRoot,
        metaRoot: metaRoot
      };
    }
  }, {
    key: "periodRoot",
    value: function periodRoot() {
      var _this$periodData = this.periodData(),
          periodRoot = _this$periodData.periodRoot;

      return periodRoot;
    }
  }, {
    key: "prevPeriodProof",
    value: function prevPeriodProof() {
      if (!this.usePrev) {
        throw Error('not set to use prev period in proofs');
      }

      var _this$periodData2 = this.periodData(),
          casRoot = _this$periodData2.casRoot;

      var proof = []; // fees hash

      proof.push((0, _ethereumjsUtil.bufferToHex)(Buffer.alloc(32, 0)));
      proof.push(this.prevHash);
      proof.push(this.merkleRoot());
      proof.push((0, _ethereumjsUtil.bufferToHex)(casRoot));
      return proof;
    } // returns [currentPeriodRoot, prevPeriodRoot]

  }, {
    key: "proof",
    value: function proof(tx) {
      var periodPos = -1;

      for (var i = 0; i < this.blockList.length; i++) {
        var pos = this.blockList[i].txHashList.indexOf(tx.hash());

        if (pos >= 0) {
          if (periodPos >= 0) {
            throw Error('tx doublespend');
          }

          periodPos = i;
        }
      }

      if (periodPos < 0) {
        throw Error('tx not in this period');
      } // get block proof in period


      var blockProof = this.getMerkleTree().getProof(this.blockList[periodPos].getMerkleTree().getRoot()).map(function (buff) {
        return (0, _ethereumjsUtil.bufferToHex)(buff);
      }); // get tx proof in block

      var proof = this.blockList[periodPos].proof(tx, periodPos); // merge blocks proof into tx proof

      blockProof.forEach(function (elem) {
        return proof.push(elem);
      }); // add extensible proof structure

      var _this$periodData3 = this.periodData(),
          casRoot = _this$periodData3.casRoot,
          periodRoot = _this$periodData3.periodRoot,
          metaRoot = _this$periodData3.metaRoot; // update the proof


      proof.push((0, _ethereumjsUtil.bufferToHex)(metaRoot));
      proof.push((0, _ethereumjsUtil.bufferToHex)(casRoot));
      proof[0] = periodRoot;
      return proof;
    }
    /**
     * Returns the block number interval for the period given block is included to.
     *
     * @param {Number} blockNumber block height of the block we are getting period block range for
     * @returns {Array} block interval in [startBlock: number, endBlock: number] format
     */

  }], [{
    key: "evaluatePrevPeriodProof",
    value: function evaluatePrevPeriodProof(proof) {
      var result;
      result = Buffer.alloc(64, 0); // fees hash

      (0, _ethereumjsUtil.toBuffer)(proof[0]).copy(result); // prevPeriod

      (0, _ethereumjsUtil.toBuffer)(proof[1]).copy(result, 32);
      var metaRoot = (0, _ethereumjsUtil.keccak256)(result);
      result = Buffer.alloc(64, 0); // blocks root

      (0, _ethereumjsUtil.toBuffer)(proof[2]).copy(result);
      metaRoot.copy(result, 32);
      var consensusRoot = (0, _ethereumjsUtil.keccak256)(result);
      result = Buffer.alloc(64, 0);
      (0, _ethereumjsUtil.toBuffer)(consensusRoot).copy(result);
      (0, _ethereumjsUtil.toBuffer)(proof[3]).copy(result, 32);
      return [proof[1], (0, _ethereumjsUtil.bufferToHex)((0, _ethereumjsUtil.keccak256)(result))];
    }
  }, {
    key: "periodBlockRange",
    value: function periodBlockRange(blockNumber) {
      var periodNum = Math.floor(blockNumber / _constants.BLOCKS_PER_PERIOD);
      return [periodNum * _constants.BLOCKS_PER_PERIOD, (periodNum + 1) * _constants.BLOCKS_PER_PERIOD - 1];
    }
    /**
    * Creates new Period with all the blocks from the given interval.
    *
    * @param {ExtendedWeb3} plasma instance of Leap Web3 or Leap Ethers
    * @param {Number} startBlock first block to include in the period
    * @param {Number} endBlock last block to include in the period
    * @param {PeriodOpts} periodOpts options for Period object as defined in Period constructor
    * @returns {Period} period
    */

  }, {
    key: "periodForBlockRange",
    value: function periodForBlockRange(plasma, startBlock, endBlock) {
      var periodOpts = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
      return Promise.all(_util["default"].range(startBlock, endBlock).map(function (n) {
        return (plasma.eth || plasma).getBlock(n, true);
      })).then(function (blocks) {
        var blockList = blocks.filter(function (a) {
          return !!a;
        }).map(function (_ref2) {
          var number = _ref2.number,
              timestamp = _ref2.timestamp,
              transactions = _ref2.transactions;
          return _block["default"].from(number, timestamp, transactions);
        });
        return new Period(null, blockList, periodOpts);
      });
    }
    /**
    * Creates {Period} where given tx was included.
    *
    * @param {ExtendedWeb3} plasma instance of Leap Web3
    * @param {Transaction} tx transaction to create {Period} for
    * @param {PeriodOpts} periodOpts options for Period object as defined in Period constructor
    * @returns {Period} period
    */

  }, {
    key: "periodForTx",
    value: function periodForTx(plasma, tx) {
      var periodOpts = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
      var blockNumber = tx.blockNumber;

      var _Period$periodBlockRa = Period.periodBlockRange(blockNumber),
          _Period$periodBlockRa2 = _slicedToArray(_Period$periodBlockRa, 2),
          startBlock = _Period$periodBlockRa2[0],
          endBlock = _Period$periodBlockRa2[1];

      return Period.periodForBlockRange(plasma, startBlock, endBlock, periodOpts);
    }
  }]);

  return Period;
}();

exports["default"] = Period;
module.exports = exports.default;